/* -----------------------------------------------------------------------

   Copyright (C) 1998-2001,2002 Marxmeier Software AG.
   All rights reserved.

   Project:  Eloquence TurboIMAGE compatibility
   File:     hp3kexport.c
   Function: dbexport workalike using TurboIMAGE API
   $Id$

   This program is distributed in the hope that it will be usefull,
   but WITHOUT ANY WARRANTY.  No author or distributor accepts
   responsibility to anyone for the consequences of using it or for
   whether it serves any particular purpose or works at all, unless
   he says so in writing.

   We kindly ask to submit any changes to support@marxmeier.com so 
   they can be folded back into the original file.

   TODO: 
    - Add support for Business Basic decimals
    - Add support for export files > 2GB

   --------------------------------------------------
   NOTE:    Don't use stdout here !
            It may be the export target
   --------------------------------------------------

   Edit history:
   06.04.1999 - MPE porting and fixups
              - Add support for 3 digit exp file naming convention
   08.04.1999 - merged changes from Gavin
   13.11.2001 - fixed stupid typo in strdup()
   13.11.2001 - add empty key argument to DBGET() to avoid trap on MPE
   14.11.2001 - some minor corrections to make it actually
                useful on MPE
   22.11.2001 - cleanup, fix Zn/Pn conversion, add Rn conversion
   18.02.2002 - fixed exp_string()
   03.04.2002 - exp_zoned(), exp_packed() sanity checks
   15.04.2002 - merge exp_zoned(), exp_packed() changes
              - merge exp_hex_endian()
              - IMAGE limits
   25.04.2002 - HAS_LONG_LONG to handle HP C compiler limitation
                with MPE 6.0
   10.06.2002 - fix gcc3.1 warnings
   04.08.2002 - use DBOPEN mode 5, removed old passwd handling
                remove group and account from default export filenames
		MPE requires 16 bit alignment for data set names
   17.08.2002 - changed INFO202, long -> int (LP64 compliance)

----------------------------------------------------------------------- */

/* Set to 0 if intended to be used with Eloquence before A.07.00 */
#if !defined(ASSUME_ELQOUENCE_A0700)
#define ASSUME_ELQOUENCE_A0700 1
#endif

/* HP C compiler for MPE/iX 6.0 does not have long long support */
#if !defined(HAS_LONG_LONG)
#define HAS_LONG_LONG 1
#endif


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#if !defined(_WIN32)
#include <unistd.h>
#else
#include <windows.h>
#endif

#if defined(_MPEXL_SOURCE)
   #pragma intrinsic DBOPEN
   #pragma intrinsic DBINFO
   #pragma intrinsic DBXBEGIN
   #pragma intrinsic DBLOCK
   #pragma intrinsic DBFIND
   #pragma intrinsic DBGET
   #pragma intrinsic DBPUT
   #pragma intrinsic DBUPDATE
   #pragma intrinsic DBDELETE
   #pragma intrinsic DBUNLOCK
   #pragma intrinsic DBXEND
   #pragma intrinsic DBXUNDO
   #pragma intrinsic DBCLOSE
   #pragma intrinsic DBERROR
   #pragma intrinsic DBEXPLAIN
   #pragma intrinsic HPFPCONVERT
#else
#include "image3k.h"
#endif

#define TRUE  1
#define FALSE 0
#define QUOTE 0x22

#define ABS(a) ((a)<0?-(a):(a))

#ifndef BIG_ENDIAN
#define BIG_ENDIAN 0x1234
#define LITTLE_ENDIAN 0x3412
#endif

#ifndef BYTE_ORDER
#if defined(_hppa) || defined(_MPEXL_SOURCE)
#define BYTE_ORDER BIG_ENDIAN
#else
#define BYTE_ORDER LITTLE_ENDIAN
#endif
#endif


/*------------------------------------------------------------------------
      types
------------------------------------------------------------------------*/

enum ExitStatus {
   EX_OK    = 0,
   EX_ERROR = 1,
   EX_USAGE = 2,
   EX_INTR  = 3
};

#if defined(_MPEXL_SOURCE)
#define DB_NAME_SIZE 28
#define MAX_SET_CNT  199
#define MAX_ITEM_CNT 1280
#define MAX_ENTRY_LEN 5120
#else
#define DB_NAME_SIZE 256
#define MAX_SET_CNT  500
#define MAX_ITEM_CNT 2048
#define MAX_ENTRY_LEN 5120
#endif

typedef struct {
   short DbId;
   char DbName[DB_NAME_SIZE];
} IMAGEDBNAME;

#define ITEM_NAME_SIZE 16
typedef struct {
   char  Name[ITEM_NAME_SIZE];
   char  Type;
   char blank;
   short Length;
   short Count;
   short zero1, zero2;
} INFO102;

typedef struct {
   short Number;
} INFO201;

#if defined(__GNUC__)
#define PACKED __attribute__ ((packed))
#else
#define PACKED
#if defined(__hppa)
#pragma HP_ALIGN NOPADDING PUSH
#endif
#if defined(_WIN32)
#pragma pack(push, 1)
#endif
#endif

#define SET_NAME_SIZE 16
typedef struct {
   char  Name[SET_NAME_SIZE];
   char Type;
   char blank;
   short EntryLength;
   short BlockingFactor;
   short zero1, zero2;
   int Entries;
   int Capacity;
} PACKED INFO202;

#if !defined(__GNUC__)
#if defined(__hppa)
#pragma HP_ALIGN POP
#endif
#if defined(_WIN32)
#pragma pack(pop)
#endif
#endif

typedef struct {
   short Item;
   short Set;
} INFO302;


/*------------------------------------------------------------------------
      forward reference
------------------------------------------------------------------------*/

void usage(void);
void version_info(void);
void wrapup(int);
void errorhandler(const char *);

void export_set(int setno);
void export_rec(FILE *fp, char *buf);
void exp_string(FILE *fp, const char *buf, size_t size);
void exp_bin(FILE *fp, const char *buf, size_t size);
void exp_int(FILE *fp, const char *buf, size_t size);
void exp_unsigned(FILE *fp, const char *buf, size_t size);
void exp_real(FILE *fp, const char *buf, size_t size);
void exp_hp3k_real(FILE *fp, const char *buf, size_t size);
void exp_packed(FILE *fp, const char *buf, size_t size);
void exp_zoned(FILE *fp, const char *buf, size_t size);

#if defined(_MPEXL_SOURCE)

/* strdup() is not available on MPE */
char *strdup(const char *s);

#endif


/*------------------------------------------------------------------------
      globals
------------------------------------------------------------------------*/

char *argv0;

char *db_name, *db_base_name, *exp_path, *user, *passwd, *single_file;
char *exp_base_name;
int verbose, chained, restruct, auto_flag;
const char *field_sep;
FILE *s_fp;

IMAGEDBNAME db;
short mode;
short status[10];

INFO102 item[MAX_ITEM_CNT];

struct {
   short count;
   short itemno[MAX_ITEM_CNT];
} item_list;

struct {
   short count;
   short setno[MAX_SET_CNT];
} set_list;

char buf[MAX_ENTRY_LEN];


/*------------------------------------------------------------------------

   main

------------------------------------------------------------------------*/

#if defined(_WIN32)
HINSTANCE hInst;
#endif

int main(int argc, char *argv[])
{
   int c;
   extern char *optarg;
   extern int optind;
   int errflg;

   int i, len;
   INFO201 info201;
   char *p;
   char tmp_pswd[64+4];

#if defined(_WIN32)
   hInst = (HINSTANCE)GetModuleHandle(NULL);
#endif

   /* ----------------------------------------
      Welcome
     ---------------------------------------- */

   argv0 = argv[0];
   if(argc < 2 || !strcmp(argv[1], "-help"))
      usage();
   if(argc > 1 && !strcmp(argv[1], "-version"))
      version_info();

   /* ----------------------------------------
      handle commandline args

      If default_to_os_user is defined
      we pre-set the user with the login
      name (not available on MPE).
      Otherwise we default to UCL 0 (on MPE)
      or the "public" user on Eloquence.
      on MPE the user name is passed as
      the USERIDENT if specified.
     ---------------------------------------- */

   user = passwd = NULL;
   errflg = FALSE;
   while((c = getopt(argc,argv, "u:cvo:p:rs:af:")) != EOF)
      switch(c) {
         case 'u':
            user = optarg;
            break;
         case 'p':
            passwd = optarg;
            break;
         case 'o':               /* set output path */
            exp_path = optarg;
            break;
         case 'c':               /* chained export mode */
            chained = TRUE;
            break;
         case 'v':               /* verbose mode */
            verbose++;
            break;
         case 'r':               /* output restructure information */
            restruct = TRUE;
            break;
         case 's':               /* output to single file */
            single_file = optarg;
            break;
         case 'a':               /* export auto dataset */
            auto_flag = TRUE;
            break;
         case 'f':               /* field separator */
            field_sep = optarg;
            break;
         default:                /* error */
            errflg = TRUE;
      }

   if(errflg) {
      fprintf(stderr, "%s: %s\n", argv0, "Illegal arguments");
      exit(EX_USAGE);
   }
   if(optind == argc) {
      fprintf(stderr, "%s: %s\n", argv0, "No database specified");
      exit(EX_USAGE);
   }

   if( ! field_sep || *field_sep == '\0' )
      field_sep = ",";

   db_name = argv[optind];
   if(*db_name == '\0') {
      fprintf(stderr, "%s: %s\n", argv0, "Illegal database name");
      exit(EX_ERROR);
   }

   p = strrchr(db_name, '/');
   if(!p)
      db_base_name = strdup(db_name);
   else
      db_base_name = strdup(p+1);
   if(db_base_name == NULL) {
      fprintf(stderr, "%s: %s\n", argv0, "Memory allocation failed");
      exit(EX_ERROR);
   }

   p = db_base_name;
   while(*p) {
      *p = toupper(*(unsigned char *)p);
      p++;
   }

   exp_base_name = strdup(db_base_name);
   if(exp_base_name == NULL) {
      fprintf(stderr, "%s: %s\n", argv0, "Memory allocation failed");
      exit(EX_ERROR);
   }

#if defined(_MPEXL_SOURCE)
   /* on MPE, remove group and account from database name. Otherwise
      they will end up in the default export filename and may cause
      problems because file names exceed 16 characters */ 
   p = strchr(exp_base_name, '.');
   if(p)
      *p = '\0';
#endif

   if(exp_path == NULL)
      exp_path = ".";
   if(passwd == NULL)
      passwd = "";

   /* If no user name (eqiv. to USERIDENT) is specified we only
      provide the password argument. On Eloquence this is
      ignored (and identical to using "public") as we always
      need a user/password pair but it makes a difference on MPE */

   if(user)
      sprintf(tmp_pswd, "%.32s;/%.32s;", passwd, user);
   else
      sprintf(tmp_pswd, "%.32s;", passwd);


   /* ----------------------------------------
      connect database
     ---------------------------------------- */

   if(verbose) {
      fprintf(stderr,
            "%s : %s\n", "Processing database", db_name);
      if(single_file == NULL)
         fprintf(stderr,
            "%s : %s\n", "Export path        ", exp_path);
      else
         fprintf(stderr,
            "%s : %s\n", "Export file        ",
               !strcmp(single_file, "-") ? "stdout" : single_file);
      fprintf(stderr, "\n");
   }

   len = strlen(db_name);
   if(len >= DB_NAME_SIZE) {
      fprintf(stderr, "%s: %s\n", db_name, "Bad database name");
      exit(EX_ERROR);
   }
   memset(&db.DbId, ' ', 2);
   memcpy(db.DbName, db_name, len);
   db.DbName[len] = ';';

   mode = 5;
   DBOPEN(&db, tmp_pswd, &mode, status);
   errorhandler("opening database");


   /* ----------------------------------------
      open single_file
     ---------------------------------------- */

   if(single_file != NULL) {
      if(!strcmp(single_file, "-"))
         s_fp = stdout;
      else {
         if((s_fp = fopen(single_file, "w")) == NULL) {
            perror(single_file);
            wrapup(EX_ERROR);
         }
      }

      fprintf(s_fp, "# DATABASE = %s\n", db_base_name);
   }

   /* ----------------------------------------
      export
     ---------------------------------------- */

   if(verbose) {
      fprintf(stderr,
            "%s\n", "DATA SET               RECORDS  COUNT");
      fprintf(stderr,
            "%s\n", "---------------- --- - -------- --------");
   }

   if(argc == optind+1) {
      mode = 203;
      DBINFO(&db, "", &mode, status, &set_list);
      errorhandler("info 203");
      for (i = 0 ; i < set_list.count; i++)
         export_set(set_list.setno[i]);
   }
   else {
      for(i = optind+1; i < argc; i++) {
         if(isdigit(*(unsigned char *)argv[i])) {
            export_set(atoi(argv[i]));
	 }
         else {
	    /* on MPE 16 bit alignment is needed even for data set names. */
	    char tmp_dset_name[16+1+1];

            if(strlen(argv[i]) > 16) {
               fprintf(stderr, "%s: %s\n", argv[i], "Illegal data set name");
               errflg++;
               continue;
	    }
            sprintf(tmp_dset_name, "%s;", argv[i]);

            mode = 201;
            DBINFO(&db, tmp_dset_name, &mode, status, &info201);
            if(status[0] == -21) {
               fprintf(stderr, "%s: %s\n", argv[i], "Illegal data set name");
               errflg++;
               continue;
            }
            errorhandler("info 201");
            export_set(info201.Number);
         }
      }
   }

   wrapup(errflg ? EX_ERROR:EX_OK);
   return( EX_OK );                              /* avoid compiler warning */
}


/*------------------------------------------------------------------------
      usage
------------------------------------------------------------------------*/

void usage()
{
   static char *msg[] = {
      "usage: DBEXPORT [options] database [set ...]",
      "options:",
      " -help    - show usage (this list)",
      " -version - show version information",
      " -u user  - set user name",
      " -p pswd  - set password",
      " -o path  - set output directory (not single file)",
      " -v       - verbose output",
      " -c       - chained export",
      " -a       - export automatic sets",
      " -r       - create restructure information",
      " -s file  - output into single file, '-' = stdout",
      " -f sep   - field separator, default is ','",
      "<$END>"
   };

   int i;
   char *p;

   for(i = 0; strcmp((p = msg[i]), "<$END>"); i++)
      fprintf(stderr, "%s\n", p);
   fprintf(stderr, "\n");
   exit(EX_USAGE);
}

void version_info()
{
   static char *msg[] = {
      "DBEXPORT (Eloquence MPE tools) 2002-08-17",
      "Copyright 1998-2001,2002 Marxmeier Software AG",
      "<$END>"
   };

   int i;
   char *p;

   printf("%s (" __DATE__ ")\n", msg[0]);
   for(i = 1; strcmp((p = msg[i]), "<$END>"); i++)
      printf("%s\n", p);
   printf("\n");
   exit(EX_OK);
}


/*------------------------------------------------------------------------
      close data base
------------------------------------------------------------------------*/

void wrapup(int cond)
{
   static int once;
   
   if(!once++) {
      short mode = 1;
      DBCLOSE(&db, "", &mode, status);
      if(status[0] != -11)
         errorhandler("closing database");
      exit(cond);
   }
}


/*------------------------------------------------------------------------
      process image status
------------------------------------------------------------------------*/

void errorhandler(const char *action)
{
   char tmp[80];
   short textlen;

   if(status[0])
   {
      fprintf(stderr,
            "Fatal image problem #%d while %s", status[0], action);
      fprintf(stderr, "\n");

      /* DBEXPLAIN(status); */
      DBERROR(status, tmp, &textlen);
      fprintf(stderr, "%.*s\n", textlen, tmp);
      wrapup(EX_ERROR);
   }
}


/*------------------------------------------------------------------------
      export_set() - export data set
------------------------------------------------------------------------*/

void export_set(int setno)
{
   INFO202 set;
   INFO302 info302;
   short s_setno, s_itemno, s_msetno;

   FILE *fp;
   int i, cnt;
   char fname[1024+1];

   /* ------------------------------
      set info
     ------------------------------ */

   s_setno = ABS(setno);
   mode = 202;
   DBINFO(&db, &s_setno, &mode, status, &set);
   errorhandler("info 202");

   if(verbose)
      fprintf(stderr,
            "%16.16s %03d %c %8d ",
            set.Name, s_setno, set.Type, set.Entries);

   if(   (set.Type == 'A' && !auto_flag)
      || (set.Entries == 0 && single_file == NULL))
   {
      if(verbose)
         fprintf(stderr, "\n");
      return;
   }

   /* ------------------------------
      item info
     ------------------------------ */

   mode = 104;
   DBINFO(&db, &s_setno, &mode, status, &item_list);
   errorhandler("info 104");

   for(i = 0; i < item_list.count; i++) {
      s_itemno = ABS(item_list.itemno[i]);
      mode = 102;
      DBINFO(&db, &s_itemno, &mode, status, &item[i]);
      errorhandler("info 102");
   }

   /* ------------------------------
      get related master (chained)
     ------------------------------ */

   if(set.Type == 'D' && chained)
   {
      mode = 302;
      DBINFO(&db, &s_setno, &mode, status, &info302);
      errorhandler("info 302");
      s_msetno = info302.Set;
      s_itemno = info302.Item;
   }

   /* ------------------------------
      create export file
     ------------------------------ */

   if(single_file != NULL) {
      fp = s_fp;
      fprintf(fp, "\n# SET %d = %.16s\n", s_setno, set.Name);
   }
   else {
      strcpy(fname, exp_path);
      if(strlen(exp_path) && exp_path[strlen(exp_path)-1] != '/')
         strcat(fname, "/");
      sprintf(fname+strlen(fname), "%s.%03d.exp", exp_base_name, s_setno);

      if((fp = fopen(fname, "w")) == NULL) {
         if(verbose)
            fprintf(stderr, "\n");
         perror(fname);
         return;
      }
   }

   /* ------------------------------
      dump restructure info
     ------------------------------ */

   if(restruct)
   {
      int fld = 1;
      if(single_file == NULL)
         fprintf(fp, "# SET %d = %.16s\n", s_setno, set.Name);
      for(i = 0; i < item_list.count; fld += item[i++].Count)
      {
         fprintf(fp, "#  %4d %.16s", fld, item[i].Name);
         if(item[i].Count > 1)
            fprintf(fp, "%d ", item[i].Count);
         fprintf(fp, "%c%d\n", item[i].Type, item[i].Length);
      }
   }

   /* ------------------------------
      export
     ------------------------------ */

   if(set.Type != 'D' || !chained || !s_msetno)
   {
      /* serial unload */

      mode = 3;
      DBCLOSE(&db, &s_setno, &mode, status);
      errorhandler("rewind dset");

      for(cnt = 0; ; ) {
         mode = 2;
         DBGET(&db, &s_setno, &mode, status, "@;", buf, "");
         if(status[0] == 11)
            break;
         errorhandler("serial get");

         export_rec(fp, buf);
         cnt++;
         if(verbose && cnt%100 == 0)
            fprintf(stderr, "%8d\b\b\b\b\b\b\b\b", cnt);
      }
      if(verbose)
         fprintf(stderr, "%8d", cnt);
   }
   else {
      /* chained unload */

      mode = 3;
      DBCLOSE(&db, &s_msetno, &mode, status);
      errorhandler("rewind mset");

      for(cnt = 0; ; ) {
         mode = 2;
         DBGET(&db, &s_msetno, &mode, status, "@;", buf, "");
         if(status[0] == 11)
            break;
         errorhandler("serial get (master)");

         mode = 1;
         DBFIND(&db, &s_setno, &mode, status, &s_itemno, buf);
         if(status[0] != 17) {
            errorhandler("get chain head");
            for(;;) {
               mode = 5;
               DBGET(&db, &s_setno, &mode, status, "@;", buf, "");
               if(status[0] == 15)
                  break;
               errorhandler("chained get");

               export_rec(fp, buf);
               cnt++;
               if(verbose && cnt%100 == 0)
                  fprintf(stderr, "%8d\b\b\b\b\b\b\b\b", cnt);
            }
         }
      }
      if(verbose)
         fprintf(stderr, "%8d", cnt);
   }

   if(single_file == NULL)
      fclose(fp);
   if(verbose)
      fprintf(stderr, "\n");
}


/*------------------------------------------------------------------------

      export_rec() - export record

------------------------------------------------------------------------*/

void export_rec(FILE *fp, char *buf)
{
   int i, r, nbytes;

   for(i = 0; i < item_list.count; i++)
      for(r = 0; r < item[i].Count; r++)
      {
         if (i || r)
            fputc(*field_sep, fp);

         switch(item[i].Type) {
         case 'B':              /* binary, size=bytes */
            nbytes = item[i].Length;
            exp_bin(fp, buf, nbytes);
            break;

         case 'X':              /* string, size=bytes */
         case 'U':
            nbytes = item[i].Length;
            exp_string(fp, buf, nbytes);
            break;

         case 'I':              /* integer, size=halfwords */
         case 'J':
            nbytes = 2 * item[i].Length;
            exp_int(fp, buf, nbytes);
            break;

         case 'K':              /* unsigned integer, size=halfwords */
            nbytes = 2 * item[i].Length;
            exp_unsigned(fp, buf, nbytes);
            break;

         case 'E':              /* IEEE fp, size=halfwords */
            nbytes = 2 * item[i].Length;
            exp_real(fp, buf, nbytes);
            break;

         case 'R':              /* HP3K real, size=halfwords */
            nbytes = 2 * item[i].Length;
#if defined(_MPEXL_SOURCE)
            exp_hp3k_real(fp, buf, nbytes);
#else
            /* Eloquence uses IEEE format */
            exp_real(fp, buf, nbytes);
#endif
            break;

         case 'Z':              /* zoned decimal, size=bytes */
            nbytes = item[i].Length;
            exp_zoned(fp, buf, nbytes);
            break;

         case 'P':              /* packed decimal, size=nibbles */
            nbytes = item[i].Length / 2;
            exp_packed(fp, buf, nbytes);
            break;

         default:
            assert(0);
         }
         buf += nbytes;
      }
   fputc('\n', fp);
}


/*------------------------------------------------------------------------

      exp_string() - export string

------------------------------------------------------------------------*/

void exp_string(FILE *fp, const char *buf, size_t size)
{
   const char *p = buf+size;
   while(size) {
      if(*--p != ' ')
         break;
      size--;
   }

   fputc(QUOTE, fp);
   while(size--) {
      int ch = *buf++ & 0xff;
      if(ch < 0x20 || (ch >= 0x80 && ch <= 0xa0) || ch == 0xff)
         fprintf(fp, "\\%03o", ch);
      else {
         if(ch == QUOTE || ch == '\\')
            fputc('\\', fp);
         fputc(ch, fp);
      }
   }
   fputc(QUOTE, fp);
}


/*------------------------------------------------------------------------

      exp_bin() - binary

------------------------------------------------------------------------*/

void exp_bin(FILE *fp, const char *buf, size_t size)
{
   const char *p = buf+size;
   while(size) {
      if(*--p != '\0')
         break;
      size--;
   }

   fputc(QUOTE, fp);
   while(size--) {
      int ch = *buf++ & 0xff;
      fprintf(fp, "\\%03o", ch);
   }
   fputc(QUOTE, fp);
}


/*------------------------------------------------------------------------

      exp_hex_endian() - binary value to a hex number

------------------------------------------------------------------------*/

#if ASSUME_ELQOUENCE_A0700

void exp_hex_endian(FILE *fp, const char *buf, size_t size)
{
   const unsigned char *buf_p;
   int nzdgts;
   const char *hex_digits = "0123456789abcdef";

   buf_p = (unsigned char *)buf;
   fputc('x', fp);

#if BYTE_ORDER != BIG_ENDIAN
   /* little endian */
   buf_p += size;
   nzdgts = 0;
   while(size--) {
      int hn, ln, ch;

      ch = *--buf_p;
      if(ch == 0 && !nzdgts && size)
         continue;
      hn = ch >> 4;
      ln = ch & 0xf;

      fputc(hex_digits[hn], fp);
      fputc(hex_digits[ln], fp);
      nzdgts = 1;
   }
#else
   /* big endian */
   nzdgts = 0;
   while(size--) {
      int hn, ln, ch;

      ch = *buf_p++;
      if(ch == 0 && !nzdgts && size)
         continue;
      hn = ch >> 4;
      ln = ch & 0xf;

      fputc(hex_digits[hn], fp);
      fputc(hex_digits[ln], fp);
      nzdgts = 1;
   }
#endif
}

#endif	/* ASSUME_ELQOUENCE_A0700 */


/*------------------------------------------------------------------------

      exp_int() - export integer value

------------------------------------------------------------------------*/

void exp_int(FILE *fp, const char *buf, size_t size)
{
   if(size == 2)
   {
      short i2;

      memcpy(&i2, buf, sizeof(short));
      fprintf(fp, "%d", i2);
   }
   else if(size == 4)
   {
      int i4;

      memcpy(&i4, buf, sizeof(int));
      fprintf(fp, "%d", i4);
   }
#if ASSUME_ELQOUENCE_A0700
#if HAS_LONG_LONG
   else if(size == 8)
   {
      long long i8;

      memcpy(&i8, buf, sizeof(long long));
      fprintf(fp, "%lld", i8);
   }
#endif /* HAS_LONG_LONG */
   else {
      exp_hex_endian(fp, buf, size);
   }
#else
   else {
      /* Eloquence A.06.31: mapped to binary */
      exp_bin(fp, buf, size);
   }
#endif
}


/*------------------------------------------------------------------------

      exp_unsigned() - export unsigned integer value

------------------------------------------------------------------------*/

void exp_unsigned(FILE *fp, const char *buf, size_t size)
{
   if(size == 2)
   {
      unsigned short u2;

      memcpy(&u2, buf, sizeof(unsigned short));
      fprintf(fp, "%u", u2);
   }
   else if(size == 4)
   {
      unsigned int i4;

      memcpy(&i4, buf, sizeof(unsigned int));
      fprintf(fp, "%u", i4);
   }
#if ASSUME_ELQOUENCE_A0700
#if HAS_LONG_LONG
   else if(size == 8)
   {
      unsigned long long i8;

      memcpy(&i8, buf, sizeof(unsigned long long));
      fprintf(fp, "%llu", i8);
   }
#endif /* HAS_LONG_LONG */
   else {
      exp_hex_endian(fp, buf, size);
   }
#else
   else {
      /* Eloquence A.06.31: mapped to binary */
      exp_bin(fp, buf, size);
   }
#endif
}


/*------------------------------------------------------------------------

      exp_real() - export IEEE floating point

------------------------------------------------------------------------*/

void exp_real(FILE *fp, const char *buf, size_t size)
{
   if(size == 4)
   {
      float f;

      memcpy(&f, buf, sizeof(float));
      fprintf(fp, "%1.7g", f);
   }
   else if(size == 8)
   {
      double d;

      memcpy(&d, buf, sizeof(double));
      fprintf(fp, "%1.15g", d);
   }
#if ASSUME_ELQOUENCE_A0700
   else {
      exp_hex_endian(fp, buf, size);
   }
#else
   else {
      static int once;
      if(!once++)
         fprintf(stderr, "Ignoring unsupported E item size\n");
      fprintf(fp, "0");
   }
#endif
}


/*------------------------------------------------------------------------

      exp_hp3k_real() - export HP3000 floating point

------------------------------------------------------------------------*/

#if defined(_MPEXL_SOURCE)

/*
   On Eloquence 'R' is really 'E' and consequently
   we simply call exp_real().
*/

void exp_hp3k_real(FILE *fp, const char *buf, size_t size)
{
   double d;
   short sformat, dformat, exceptions, roundmode;
   int status;

   if(size == 4) {
      sformat = 1;	/* Hewlett-Packard 3000 32-bit */
      dformat = 4;      /* IEEE 64-bit */
      roundmode = 0;    /* Round to nearest */
      HPFPCONVERT(buf,&d,sformat,dformat,&status,&exceptions,roundmode);
      if(status)
         goto bad_number;
      fprintf(fp, "%1.7g", d);
   }
   else if(size == 8)
   {
      sformat = 2;	/* Hewlett-Packard 3000 64-bit */
      dformat = 4;      /* IEEE 64-bit */
      roundmode = 0;    /* Round to nearest */
      HPFPCONVERT(buf,&d,sformat,dformat,&status,&exceptions,roundmode);
      if(status)
         goto bad_number;
      fprintf(fp, "%1.15g", d);
   }
   else {
      static int once;
      if(!once++)
         fprintf(stderr, "Ignoring unsupported R item size\n", size);
bad_number: ;
      fprintf(fp, "0");
   }
}

#endif  /* _MPEXL_SOURCE */


/*------------------------------------------------------------------------

      exp_zoned() - export packed decimal

------------------------------------------------------------------------*/

void exp_zoned(FILE *fp, const char *buf, size_t size)
{
#if ASSUME_ELQOUENCE_A0700
   const unsigned char *buf_p;
   int sign, last_dgt, nz_dgts, idx;

   /*
      The last character in a zoned format indicates the sign in
      addition to the digit value. Unsigned values don't use a
      sign so the last character might as well be a digit.

      digit    positive negative
      -------- -------- --------
      0        {        }
      1        A        J
      2        B        K
      3        C        L
      4        D        M
      5        E        N
      6        F        O
      7        G        P
      8        H        Q
      9        I        R
   */

   buf_p = (unsigned char *)buf;
   sign = 0;				/* unsigned value */
   last_dgt = buf_p[--size];
   if(last_dgt < '0' || last_dgt > '9') {
      if(last_dgt == '{') {		/* positive, zero */
         last_dgt = '0';
	 sign = 1;
      }
      else if(last_dgt == '}') {
         last_dgt = '0';		/* negative, zero */
	 sign = -1;
      }
      else if(last_dgt >= 'A' && last_dgt <= 'I') {
         last_dgt -= 16;		/* A..I (positive) */
         sign = 1;
      }
      else if(last_dgt >= 'J' && last_dgt <= 'R') {
         last_dgt -= 25;		/* J..R (negative) */
         sign = -1;
      }
      else {
bad_number: ;
         fprintf(fp, "0");
	 return;
      }
   }
   
   /* number sanity check */

   for(idx = 0; idx < size; idx++) {
      int ch = buf_p[idx];
      if(ch < '0' || ch > '9')
         goto bad_number;
   }

   /* format digits */

   nz_dgts = 0;
   for(idx = 0; idx < size; idx++) {
      int ch = buf_p[idx];
      if(!nz_dgts) {
         if(ch == '0')
            continue;
	 nz_dgts = 1;
	 if(sign)
	    fprintf(fp, "%c", sign < 0 ? '-' : '+');
      }
      fprintf(fp, "%c", ch);
   }

   if(!nz_dgts /* && last_dgt != '0' */ && sign)
      fprintf(fp, "%c", sign < 0 ? '-' : '+');
   fprintf(fp, "%c", last_dgt);      
#else
   /* Eloquence A.06.31: mapped to binary -> string is more readable */
   exp_string(fp, buf, size);
#endif
}


/*------------------------------------------------------------------------

      exp_packed() - export packed decimal

------------------------------------------------------------------------*/

void exp_packed(FILE *fp, const char *buf, size_t size)
{
#if ASSUME_ELQOUENCE_A0700
   const unsigned char *buf_p;
   int sign, sign_nbl, last_dgt, nz_dgts, idx;

   /*
      Nibble packed BCD format. The last nibble indicates the sign
      or an indicator for an unsigned values.

      sign     meaning
      -------- -----------------
      0xc      positive value
      0xd      negative value
      0xf      unsigned value
   */

   buf_p = (unsigned char *)buf;
   size--;
   sign_nbl = buf_p[size] & 0xf;
   last_dgt = buf_p[size] >> 4;

   if(sign_nbl == 0xc)				/* positive */
      sign = 1;
   else if(sign_nbl == 0xd)			/* negative */
      sign = -1;
   else	if(sign_nbl == 0xf)			/* unsigned */
      sign = 0;
   else {
bad_number: ;
      fprintf(fp, "0");
      return;
   }

   /* number sanity check */

   if(last_dgt > 9)
      goto bad_number;
   for(idx = 0; idx < size; idx++) {
      int ch = buf_p[idx];
      if(((ch & 0xf0) > 0x90) || ((ch & 0xf) > 9))
         goto bad_number;
   }

   /* format digits */

   nz_dgts = 0;
   for(idx = 0; idx < size; idx++) {
      int hn, ln, ch;

      ch = buf_p[idx];
      hn = ch >> 4;
      ln = ch & 0xf;

      if(!nz_dgts) {
         if(!ch)
            continue;
	 nz_dgts = 1;
	 if(sign)
	    fprintf(fp, "%c", sign < 0 ? '-' : '+');
	 if(hn)
	    fprintf(fp, "%c", '0' + hn);
	 fprintf(fp, "%c", '0' + ln);
      }
      else {
         fprintf(fp, "%c", '0' + hn);
	 fprintf(fp, "%c", '0' + ln);
      }
   }

   if(!nz_dgts /* && last_dgt */ && sign)
      fprintf(fp, "%c", sign < 0 ? '-' : '+');
   fprintf(fp, "%c", '0' + last_dgt);      
#else
   /* Eloquence A.06.31: mapped to binary */
   exp_bin(fp, buf, size);
#endif
}


/*------------------------------------------------------------------------

      porting

------------------------------------------------------------------------*/

#if defined(_MPEXL_SOURCE)

/* strdup() is not available on MPE */

char *strdup(const char *s)
{
   char *p = malloc( strlen(s)+1 );
   if(p)
      strcpy(p, s);
   return p;
}

#endif  /* _MPEXL_SOURCE */
