/* -----------------------------------------------------------------------

   (C) Copyright Marxmeier Software AG 1998-2001,2002. 
   All rights reserved.

   Project:  Eloquence TurboIMAGE compatibility
   File:     hp3kinfo.c
   Function: Simple database information using TurboIMAGE API
   $Id$

   This program is distributed in the hope that it will be usefull,
   but WITHOUT ANY WARRANTY.  No author or distributor accepts
   responsibility to anyone for the consequences of using it or for
   whether it serves any particular purpose or works at all, unless
   he says so in writing.

   We kindly ask to submit any changes to support@marxmeier.com so 
   they can be folded back into the original file.

   Edit history:
   15.04.2002 - IMAGE limits
   10.06.2002 - fix gcc3.1 warnings
   31.07.2002 - changed MAX_SET_CNT 199 -> 240   
   04.08.2002 - MPE requires 16 bit alignment for data set names
   17.08.2002 - changed INFO202, long -> int (LP64 compliance)

----------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#if !defined(_WIN32)
#include <unistd.h>
#else
#include <windows.h>
#endif

#if defined(_MPEXL_SOURCE)
   #pragma intrinsic DBOPEN
   #pragma intrinsic DBINFO
   #pragma intrinsic DBXBEGIN
   #pragma intrinsic DBLOCK
   #pragma intrinsic DBFIND
   #pragma intrinsic DBGET
   #pragma intrinsic DBPUT
   #pragma intrinsic DBUPDATE
   #pragma intrinsic DBDELETE
   #pragma intrinsic DBUNLOCK
   #pragma intrinsic DBXEND
   #pragma intrinsic DBXUNDO
   #pragma intrinsic DBCLOSE
   #pragma intrinsic DBERROR
   #pragma intrinsic DBEXPLAIN
#else
#include "image3k.h"
#endif

#define TRUE 1
#define FALSE 0

#define ABS(a) ((a)<0?-(a):(a))


/*------------------------------------------------------------------------
      types
------------------------------------------------------------------------*/

enum ExitStatus {
   EX_OK    = 0,
   EX_ERROR = 1,
   EX_USAGE = 2,
   EX_INTR  = 3
};

#if defined(_MPEXL_SOURCE)
#define DB_NAME_SIZE 28
#define MAX_SET_CNT  240
#else
#define DB_NAME_SIZE 256
#define MAX_SET_CNT  500
#endif

typedef struct {
   short DbId;
   char DbName[DB_NAME_SIZE];
} IMAGEDBNAME;

typedef struct {
   short Number;
} INFO201;

#if defined(__GNUC__)
#define PACKED __attribute__ ((packed))
#else
#define PACKED
#if defined(__hppa)
#pragma HP_ALIGN NOPADDING PUSH
#endif
#if defined(_WIN32)
#pragma pack(push, 1)
#endif
#endif

#define SET_NAME_SIZE 16
typedef struct {
   char  Name[SET_NAME_SIZE];
   char Type;
   char blank;
   short EntryLength;
   short BlockingFactor;
   short zero1, zero2;
   int Entries;
   int Capacity;
} PACKED INFO202;

#if !defined(__GNUC__)
#if defined(__hppa)
#pragma HP_ALIGN POP
#endif
#if defined(_WIN32)
#pragma pack(pop)
#endif
#endif

char *argv0;
char *db_name, *user, *passwd;

IMAGEDBNAME db;
short mode;
short status[10];

void usage(), version_info(), wrapup(), errorhandler(), set_info();


/*------------------------------------------------------------------------
      main
------------------------------------------------------------------------*/

#if defined(_WIN32)
HINSTANCE hInst;
#endif

int main(int argc, char *argv[])
{
   int c;
   extern char *optarg;
   extern int optind;
   int errflg;

   int i, len;
   INFO201 info201;
   struct {
      short count;
      short setno[MAX_SET_CNT];
   } set_list;

   char tmp_pswd[64+4];

#if defined(_WIN32)
   hInst = (HINSTANCE)GetModuleHandle(NULL);
#endif

   /* ----------------------------------------
      welcome
     ---------------------------------------- */

   argv0 = argv[0];
   if(argc < 2 || !strcmp(argv[1], "-help"))
      usage();
   if(argc > 1 && !strcmp(argv[1], "-version"))
      version_info();

   /* ----------------------------------------
      handle commandline args
     ---------------------------------------- */

   user = passwd = NULL;
   errflg = FALSE;
   while((c = getopt(argc,argv, "u:p:")) != EOF)
      switch(c) {
         case 'u':
            user = optarg;
            break;
         case 'p':
            passwd = optarg;
            break;
         default:
            errflg = TRUE;
      }

   if(optind == argc) {
      fprintf(stderr, "%s: %s\n", argv[0], "No database specified");
      errflg++;
   }

   if(errflg)
      usage();

   /* ----------------------------------------
      connect database
     ---------------------------------------- */

   db_name = argv[optind++];

   if(passwd == NULL)
      passwd = "";
   if(user)
      sprintf(tmp_pswd, "%.32s;/%.32s;", passwd, user);
   else
      sprintf(tmp_pswd, "%.32s;", passwd);
   
   printf("%s : %s\n\n", "Processing database", db_name);

   len = strlen(db_name);
   if(len >= DB_NAME_SIZE) {
      fprintf(stderr, "%s: %s\n", db_name, "Bad database name");
      exit(EX_ERROR);
   }
   memset(&db.DbId, ' ', 2);
   memcpy(db.DbName, db_name, len);
   db.DbName[len] = ';';

   mode = 5;
   DBOPEN(&db, tmp_pswd, &mode, status);
   errorhandler("opening database");


   /* ------------------------------
      header
     ------------------------------ */

   printf("%s\n", "SET NAME              RECLEN CAPACITY ENTRIES");
   printf("%s\n", "---------------- -- - ------ -------- --------");


   /* ------------------------------
      info
     ------------------------------ */

   if(argc == optind) {
      mode = 203;
      DBINFO(&db, "", &mode, status, &set_list);
      errorhandler("info 203");
      for (i = 0 ; i < set_list.count; i++)
         set_info(set_list.setno[i]);
   }
   else {
      for(i = optind; i < argc; i++) {
         if(isdigit(*(unsigned char *)argv[i]))
            set_info(atoi(argv[i]));
         else {
	    /* on MPE 16 bit alignment is needed even for data set names. */
	    char tmp_dset_name[16+1+1];

            if(strlen(argv[i]) > 16) {
               fprintf(stderr, "%s: %s\n", argv[i], "Illegal data set name");
               errflg++;
               continue;
	    }
            sprintf(tmp_dset_name, "%s;", argv[i]);

	    mode = 201;
            DBINFO(&db, tmp_dset_name, &mode, status, &info201);
            if(status[0] == -21) {
               fprintf(stderr, "%s: %s\n", argv[i], "Illegal set name");
               errflg++;
               continue;
            }
            errorhandler("info 201");
            set_info(info201.Number);
         }
      }
   }
   
   wrapup(errflg ? EX_ERROR:EX_OK);
   return( EX_OK );                              /* avoid compiler warning */
}


/*------------------------------------------------------------------------
      usage
------------------------------------------------------------------------*/

void usage()
{
   static char *msg[] = {
      "usage: DBINFO [options] database [set ...]",
      "options:",
      " -help    - show usage (this list)",
      " -version - show version information",
      " -u user  - set user name",
      " -p pswd  - set password",
      "<$END>"
   };

   int i;
   char *p;

   for(i = 0; strcmp((p = msg[i]), "<$END>"); i++)
      fprintf(stderr, "%s\n", p);
   fprintf(stderr, "\n");
   exit(EX_USAGE);
}

void version_info()
{
   static char *msg[] = {
      "DBINFO (Eloquence MPE tools) 2002-08-17",
      "Copyright 1998-2001,2002 Marxmeier Software AG",
      "<$END>"
   };

   int i;
   char *p;

   printf("%s (" __DATE__ ")\n", msg[0]);
   for(i = 1; strcmp((p = msg[i]), "<$END>"); i++)
      printf("%s\n", p);
   printf("\n");
   exit(EX_OK);
}


/*------------------------------------------------------------------------
      set info
------------------------------------------------------------------------*/

void set_info(int setno)
{
   INFO202 info202;
   short ssetno;
   
   ssetno = ABS(setno);
   mode = 202;
   DBINFO(&db, &ssetno, &mode, status, &info202);
   errorhandler("info 202");
   printf("%16.16s %02d %c %6d %8d %8d",
            info202.Name, ssetno, info202.Type, info202.EntryLength,
            info202.Capacity, info202.Entries);
   printf("\n");
}


/*------------------------------------------------------------------------
      close data base
------------------------------------------------------------------------*/

void wrapup(int cond)
{
   static int once;
   
   if(!once++) {
      short mode = 1;
      DBCLOSE(&db, "", &mode, status);
      if(status[0] != -11)
         errorhandler("closing database");
      exit(cond);
   }
}


/*------------------------------------------------------------------------
      process image status
------------------------------------------------------------------------*/

void errorhandler(const char *action)
{
   char tmp[80];
   short textlen;
   
   if(status[0])
   {
      fprintf(stderr,
            "Fatal image problem #%d while %s", status[0], action);
      fprintf(stderr, "\n");

      /* DBEXPLAIN(status); */
      DBERROR(status, tmp, &textlen);
      fprintf(stderr, "%.*s\n", textlen, tmp);
      wrapup(EX_ERROR);
   }
}
